import cv2
import json
import pandas as pd
import os
import datetime
from qreader import QReader
from pathlib import Path

class QRScanner:
    def __init__(self, source=0, output_folder="qr_output"):
        self._source = source
        self._qreader = QReader()
        self._seen_qr_codes = set()
        self._output_folder = output_folder
        os.makedirs(self._output_folder, exist_ok=True)

        timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self._json_file = os.path.join(self._output_folder, f"qr_data_{timestamp}.json")
        self._excel_file = os.path.join(self._output_folder, f"qr_data_{timestamp}.xlsx")
        
        self._capture = None

    def start(self):
        """Starts the QR scanning process based on the source type."""
        try:
            if isinstance(self._source, int):
                self._start_webcam()
            elif isinstance(self._source, str):
                self._process_source()
            else:
                raise ValueError("❌ Invalid source type.")
        finally:
            self._cleanup()

    def _start_webcam(self):
        print("📹 Starting webcam capture... Press 'q' to exit.")
        self._capture = cv2.VideoCapture(self._source)
        if not self._capture.isOpened():
            raise ValueError("❌ Could not open webcam")

        while True:
            ret, frame = self._capture.read()
            if not ret:
                break

            self._detect_qr(frame)
            cv2.imshow("QR Code Scanner", frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break

    def _process_source(self):
        source_path = Path(self._source)

        if "http" in self._source:
            self._start_rtsp()
        elif source_path.is_dir():
            self._process_folder(source_path)
        elif source_path.is_file():
            self._process_file(source_path)
        else:
            raise ValueError("❌ Invalid source path")

    def _start_rtsp(self):
        print("🎥 Starting RTSP stream... Press 'q' to exit.")
        self._capture = cv2.VideoCapture(self._source)
        if not self._capture.isOpened():
            raise ValueError("❌ Could not open RTSP stream")

        while True:
            ret, frame = self._capture.read()
            if not ret:
                break

            self._detect_qr(frame)
            cv2.imshow("QR Code Scanner", frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break

    def _process_folder(self, folder_path):
        print(f"\n📁 Processing folder: {folder_path}")
        for file_path in Path(folder_path).iterdir():
            self._process_file(file_path)

    def _process_file(self, file_path):
        if file_path.suffix.lower() in {'.jpg', '.jpeg', '.png', '.bmp', '.tiff', '.webp'}:
            self._process_image(file_path)
        elif file_path.suffix.lower() in {'.mp4', '.avi', '.mov', '.mkv', '.wmv', '.flv'}:
            self._process_video(file_path)
        else:
            print(f"⚠️ Unsupported file type: {file_path}")

    def _process_image(self, image_path):
        print(f"\n📸 Processing image: {image_path}")
        frame = cv2.imread(str(image_path))
        if frame is not None:
            self._detect_qr(frame)

    def _process_video(self, video_path):
        print(f"\n🎥 Processing video: {video_path}")
        self._capture = cv2.VideoCapture(str(video_path))
        if not self._capture.isOpened():
            print(f"⚠️ Error opening video: {video_path}")
            return

        while True:
            ret, frame = self._capture.read()
            if not ret:
                break

            self._detect_qr(frame)
            cv2.imshow("QR Code Scanner", frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break

        self._capture.release()

    def _detect_qr(self, frame):
        qr_results = self._qreader.detect_and_decode(frame)
        new_qr_codes = {qr.strip() for qr in qr_results if qr} - self._seen_qr_codes

        if new_qr_codes:
            self._seen_qr_codes.update(new_qr_codes)
            print("\n✅ New QR Codes Detected:")
            for qr_data in new_qr_codes:
                print(qr_data)

            self._save_qr_data()

    def _save_qr_data(self):
        if not self._seen_qr_codes:
            return

        qr_list = [self._parse_qr_text(qr) for qr in self._seen_qr_codes]

        with open(self._json_file, "w") as f:
            json.dump(qr_list, f, indent=4)

        df = pd.DataFrame(qr_list).dropna(how='all')
        df.to_excel(self._excel_file, index=False)

        print(f"\n📂 QR data saved: {len(qr_list)} codes")

    def _parse_qr_text(self, qr_text):
        qr_data = {}
        for line in qr_text.split("\n"):
            parts = line.split(":", 1)
            if len(parts) == 2:
                qr_data[parts[0].strip()] = parts[1].strip()
        return qr_data

    def _cleanup(self):
        if self._capture is not None:
            self._capture.release()
        cv2.destroyAllWindows()
